(function(){
  var re_let_varname, re_macro_fn_name, re_macro_arg_names, re_for;
  function escape(str){
    return str.replace(/\\/g, '\\\\').replace(/["']/g, function(str){
      return "\\" + str;
    }).replace('\n', '\\n');
  }
  function trim(str){
    return str.replace(/^\s*/, "").replace(/\s*$/, "");
  }
  function BasicNode(specs, that){
    specs == null && (specs = {});
    that == null && (that = {});
    that.specs = specs;
    that.compile = (function(){
      function compile(opts, ctx){
        return "_res += '" + specs.contents + "';";
      }
      return compile;
    }());
    return that;
  }
  function NodeList(specs, that){
    var nodes, _last_is_string;
    BasicNode(specs, that);
    _last_is_string = false;
    nodes = [];
    that.compile = (function(){
      function compile(opts, ctx){
        var i, res, node, _i, _ref, _len;
        ctx = ctx || [];
        i = 0;
        res = "";
        for (_i = 0, _len = (_ref = nodes).length; _i < _len; ++_i) {
          node = _ref[_i];
          res += node.compile(opts, ctx);
        }
        return res;
      }
      return compile;
    }());
    that.push = (function(){
      function push(node){
        if (node.charAt) {
          if (_last_is_string) {
            nodes[nodes.length - 1].specs.contents += escape(node);
          } else {
            nodes.push(BasicNode({
              contents: escape(node)
            }));
          }
          return _last_is_string = true;
        } else {
          _last_is_string = false;
          return nodes.push(node);
        }
      }
      return push;
    }());
    that.nodes = nodes;
    return that;
  }
  function NodePrint(specs, that){
    BasicNode(specs, that);
    that.compile = (function(){
      function compile(opts, ctx){
        return "_res += (" + make_expr(specs.contents, ctx) + ") || '';";
      }
      return compile;
    }());
    return that;
  }
  re_let_varname = /\s*([a-zA-Z_]\w+)\s*=/;
  function NodeLet(specs, that){
    var match, var_name;
    BasicNode(specs, that);
    match = re_let_varname.exec(specs.contents);
    if (match === null) {
      throw "Let was expecting a correct variable name";
    }
    var_name = match[1];
    that.compile = (function(){
      function compile(opts, ctx){
        ctx.push(var_name);
        return specs.contents + ";";
      }
      return compile;
    }());
    return that;
  }
  function NodeTag(specs, that){
    BasicNode(specs, that);
    that.until = "none";
    return that;
  }
  re_macro_fn_name = /\s*([a-zA-Z_]\w+)\s*\((.*)\)/;
  re_macro_arg_names = /[a-zA-Z_]\w+/g;
  function NodeMacro(specs, that){
    var match, fn_name, var_names;
    NodeTag(specs, that);
    that.until = "endtag";
    match = re_macro_fn_name.exec(specs.contents);
    if (match === null) {
      throw "Macro declaration is invalid";
    }
    fn_name = match[1];
    var_names = [];
    match[2].replace(re_macro_arg_names, function(s){
      var_names.push(s);
      return s;
    });
    that.compile = (function(){
      function compile(opts, ctx){
        var new_ctx, variable_name, that, _res, _i, _ref, _len;
        _res = "";
        ctx.push(fn_name);
        new_ctx = ctx.slice(0);
        for (_i = 0, _len = (_ref = var_names).length; _i < _len; ++_i) {
          variable_name = _ref[_i];
          new_ctx.push(variable_name);
        }
        _res += "function " + specs.contents + " { var _res = '';";
        if (that = that.inside) {
          _res += that.inside.compile(opts, new_ctx);
        }
        _res += "return _res; }";
        return _res;
      }
      return compile;
    }());
    return that;
  }
  function NodeElse(specs, that){
    NodeTag(specs, that);
    that.compile = (function(){
      function compile(opts, ctx){
        return "} else { ";
      }
      return compile;
    }());
    return that;
  }
  function NodeIf(specs, that){
    var condition;
    NodeTag(specs, that);
    that.until = "endtag";
    condition = specs.contents;
    that.compile = (function(){
      function compile(opts, ctx){
        var res, that;
        res = "if (" + make_expr(condition, ctx) + " {";
        if (that = that.inside) {
          res += that.inside.compile(opts, ctx);
        }
        res += "} ";
        return res;
      }
      return compile;
    }());
    return that;
  }
  function NodeDo(specs, that){
    NodeTag(specs, that);
    that.compile = (function(){
      function compile(opts, ctx){
        return make_expr(specs.contents) + ";";
      }
      return compile;
    }());
    that.until = "none";
    return that;
  }
  function NodeElseIf(specs, that){
    var condition;
    NodeIf(specs, that);
    condition = specs.contents;
    that.compile = (function(){
      function compile(opts, ctx){
        return " } else if (" + make_expr(condition, ctx) + ") {";
      }
      return compile;
    }());
    that.until = "none";
    return that;
  }
  re_for = /([a-zA-Z_]\w*)\s*(,\s*([a-zA-Z_]\w*))?\s*in\s*(.*)/;
  function NodeFor(specs, that){
    var m, iterator, value_iterator, container_expression;
    NodeTag(specs, that);
    that.until = "endtag";
    m = re_for.exec(specs.contents);
    if (m === null) {
      throw "For expression in invalid";
    }
    iterator = m[1];
    value_iterator = m[3];
    container_expression = m[4];
    that.compile = (function(){
      function compile(opts, ctx){
        var container, res, that;
        opts == null && (opts = {});
        ctx = ctx.slice(0);
        ctx.push(iterator);
        if (value_iterator) {
          ctx.push(value_iterator);
        }
        ctx.push('loop');
        container = make_expr(trim(container_expression, ctx));
        res = "";
        if (that = value_iterator === undefined) {
          res += "(function () {                                  i = 0;var " + iterator + " = null;var __expr = " + container + " || [];var loop = {};loop.length = __expr.length;loop.callee = arguments.callee;for (i = 0; i < __expr.length; i++) {loop.index = i + 1; loop.index0 = i;loop.first = i == 0; loop.last = i == __expr.length - 1;" + iterator + " = __expr[i];";
          if (that = that.inside) {
            res += that.inside.compile(opts, ctx);
          }
          res += "} }) ();";
        } else {
          res += "(function () {x = 0;i = 0;var " + iterator + " = null;var " + iterator_value + " = null;var __expr = " + container + " || {};var loop = {};loop.callee = arguments.callee;for (x in __expr) {loop.index = i + 1; loop.index0 = i;i += 1;" + iterator + " = x;" + iterator_value + " = __expr[x];";
          if (that = that.inside) {
            res += that.inside.compile(opts, ctx);
          }
          res += "} }) ();";
        }
        return res;
      }
      return compile;
    }());
    return that;
  }
  exports.nodes = {
    'if': NodeIf,
    'else': NodeElse,
    'elseif': NodeElseIf,
    'do': NodeDo,
    'let': NodeLet,
    'for': NodeFor,
    'macro': NodeMacro
  };
}).call(this);
